@@ -13,7 +13,11 @@ except ImportError:
     from tqdm import tqdm
 
 from hashlib import md5
+from pathlib import Path
 from copy import deepcopy
+from filetype import guess
+from flatten_dict import flatten, unflatten
+from base64 import b64encode, b64decode
 from urllib.parse import urlparse
 from pyisemail import is_email
 from collections import defaultdict
@@ -28,19 +32,29 @@ from bravado.swagger_model import Loader
 from bravado.config import bravado_config_from_config_dict
 from bravado_core.spec import Spec
 from json2html import Json2Html
-from IPython.display import display, HTML
+from IPython.display import display, HTML, Image, FileLink
 from boltons.iterutils import remap
 from pymatgen.core import Structure
 from concurrent.futures import as_completed
 from requests_futures.sessions import FuturesSession
+from filetype.types.archive import Gz
+from filetype.types.image import Jpeg, Png, Gif, Tiff
 
 
 MAX_WORKERS = 10
+MAX_ELEMS = 10
+MAX_BYTES = 200 * 1024
 DEFAULT_HOST = "contribs-api.materialsproject.org"
 BULMA = "is-narrow is-fullwidth has-background-light"
 PROVIDERS = {"github", "google", "facebook", "microsoft", "amazon"}
+COMPONENTS = {"structures", "tables", "attachments"}
 VALID_URLS = {f"http://{h}:{p}" for h in ["localhost", "contribs-api"] for p in [5000, 5002, 5003]}
 VALID_URLS |= {f"https://{n}-api.materialsproject.org" for n in ["contribs", "lightsources", "ml"]}
+VALID_URLS |= {
+    f"http://localhost.{n}-api.materialsproject.org" for n in ["contribs", "lightsources", "ml"]
+}
+SUPPORTED_FILETYPES = (Gz, Jpeg, Png, Gif, Tiff)
+SUPPORTED_MIMES = [t().mime for t in SUPPORTED_FILETYPES]
 
 j2h = Json2Html()
 pd.options.plotting.backend = "plotly"
@@ -126,6 +140,17 @@ class Dict(dict):
         )
 
 
+class Attachment(dict):
+    def pretty(self):
+        content = b64decode(self["content"], validate=True)
+
+        if self["mime"].startswith("image/"):
+            return Image(content)
+
+        Path(self["name"]).write_bytes(content)
+        return FileLink(self["name"])
+
+
 def load_client(apikey=None, headers=None, host=None):
     warnings.warn(
         "load_client(...) is deprecated, use Client(...) instead", DeprecationWarning
@@ -135,12 +160,18 @@ def load_client(apikey=None, headers=None, host=None):
 class Client(SwaggerClient):
     """client to connect to MPContribs API
 
-    We only want to load the swagger spec from the remote server when needed and not everytime the
-    client is initialized. Hence using the Borg design nonpattern (instead of Singleton): Since the
-    __dict__ of any instance can be re-bound, Borg rebinds it in its __init__ to a class-attribute
-    dictionary. Now, any reference or binding of an instance attribute will actually affect all
-    instances equally.
+    Typical usage:
+        - set environment variable MPCONTRIBS_API_KEY to the API key from your MP profile
+        - import and init:
+          >>> from mpcontribs.client import Client
+          >>> client = Client()
     """
+    # We only want to load the swagger spec from the remote server when needed and not
+    # everytime the client is initialized. Hence using the Borg design nonpattern (instead
+    # of Singleton): Since the __dict__ of any instance can be re-bound, Borg rebinds it
+    # in its __init__ to a class-attribute dictionary. Now, any reference or binding of an
+    # instance attribute will actually affect all instances equally.
+    # TODO ratelimit support in bravado (wrapped around get_entry..., monkey-patch?)
 
     _shared_state = {}
 
@@ -162,7 +193,8 @@ class Client(SwaggerClient):
         self.apikey = apikey
         self.headers = {"x-api-key": apikey} if apikey else headers
         self.host = host
-        self.protocol = "https" if host.endswith(".materialsproject.org") else "http"
+        ssl = host.endswith(".materialsproject.org") and not host.startswith("localhost.")
+        self.protocol = "https" if ssl else "http"
         self.url = f"{self.protocol}://{self.host}"
 
         if self.url not in VALID_URLS:
@@ -316,12 +348,67 @@ class Client(SwaggerClient):
             ).result()
         )
 
+    def get_attachment(self, aid_or_md5):
+        """Convenience function to get attachment.
+
+        Args:
+            aid_or_md5 (str): ObjectId or MD5 hash digest for attachment
+
+        Returns:
+            pd.DataFrame: pandas DataFrame containing table data
+        """
+        str_len = len(aid_or_md5)
+        if str_len not in {24, 32}:
+            raise ValueError(f"'{aid_or_md5}' is not a valid attachment id or md5 hash digest!")
+
+        if str_len == 32:
+            attachments = self.attachments.get_entries(
+                md5=aid_or_md5, _fields=["id"]
+            ).result()
+            if not attachments:
+                raise ValueError(f"attachment for md5 '{aid_or_md5}' not found!")
+            aid = attachments["data"][0]["id"]
+        else:
+            aid = aid_or_md5
+
+        return Attachment(self.attachments.get_entry(pk=aid, _fields=["_all"]).result())
+
     def init_columns(self, project, columns):
-        """initialize columns to set their order and desired units"""
+        """initialize columns for a project to set their order and desired units
+
+        Be aware that this function resets/overwrites the `columns` field in a project.
+        The `columns` field also tracks the minima and maxima of each `data` field as
+        contributions are submitted. This function should thus be executed before
+        submitting contributions, or all contributions for a project should be deleted to
+        ensure clean initialization of all columns. If columns are not initialized using
+        this function, `submit_contributions` will respect the order of columns as they
+        are submitted and will auto-determine suitable units based on the first
+        contribution containing a respective data column.
+
+        The `columns` argument is a dictionary which maps the data field names to its
+        units. Use `None` to indicate that a field doesn't have a unit. Nested fields are
+        indicated using a dot (".") in the data field name. Example:
+
+        >>> client.init_columns("sandbox", {"a": None, "b.c": "eV", "b.d": "mm"})
+
+        This example will result in column headers on the project landing page of the form
+
+
+        |      |      data       |
+        | data |        b        |
+        |   a  | c [eV] | d [mm] |
+
+
+        Args:
+            project (str): name of the project for which to initialize data columns
+            columns (dict): dictionary mapping data column to its unit (use None as value)
+        """
         self.projects.update_entry(pk=project, project={"columns": []}).result()
+        # sort to avoid "overlapping columns" error in handsontable's NestedHeaders
+        sorted_columns = flatten(unflatten(columns, splitter="dot"), reducer="dot")
         cols = []
 
-        for path, unit in columns.items():
+        for path, unit in sorted_columns.items():
             col = {"path": f"data.{path}"}
             if unit is not None:
                 col["unit"] = unit
@@ -332,8 +419,18 @@ class Client(SwaggerClient):
             pk=project, project={"columns": cols}
         ).result()
 
-    def delete_contributions(self, project, per_page=100, max_workers=5):
-        """Convenience function to remove all contributions for a project"""
+    def delete_contributions(self, project, per_page=100, max_workers=5, retry=False):
+        """Remove all contributions for a project
+
+        Note: This also resets the columns field for a project. It might have to be
+        re-initialized via `client.init_columns()` again.`
+
+        Args:
+            project (str): name of the project for which to delete contributions
+            per_page (int): number of contributions to delete per request
+            max_workers (int): maximum number of parallel requests to send at a time
+            retry (bool): if True, retry deletion of failed contributions until done
+        """
         tic = time.perf_counter()
 
         if max_workers > MAX_WORKERS:
@@ -342,6 +439,8 @@ class Client(SwaggerClient):
 
         cids = self.get_contributions(project)["ids"]
         total = len(cids)
+        # reset columns to be save (sometimes not all are reset BUGFIX?)
+        self.projects.update_entry(pk=project, project={"columns": []}).result()
 
         if cids:
             with FuturesSession(max_workers=max_workers) as session:
@@ -362,15 +461,21 @@ class Client(SwaggerClient):
                     self._run_futures(futures, total=len(cids))
                     cids = self.get_contributions(project)["ids"]
 
+                    if not retry:
+                        break
+
                 self.load()
 
-        # reset columns to be save (sometimes not all are reset BUGFIX?)
-        self.projects.update_entry(pk=project, project={"columns": []}).result()
-        toc = time.perf_counter()
-        dt = (toc - tic) / 60
-        print(f"It took {dt:.1f}min to delete {total} contributions.")
+            toc = time.perf_counter()
+            dt = (toc - tic) / 60
+
+            if cids:
+                print(f"There were errors and {len(cids)} contributions are left to delete!")
+            else:
+                print(f"It took {dt:.1f}min to delete {total} contributions.")
+        else:
+            print(f"There aren't any contributions to delete for {project}")
 
-    # TODO ratelimit support in bravado (wrapped around get_entry..., monkey-patch?)
     @sleep_and_retry
     @limits(calls=175, period=60)
     def get_unique_identifiers_flag(self, name):
@@ -390,6 +495,7 @@ class Client(SwaggerClient):
         ret = defaultdict(set)
         ret["unique_identifiers"] = self.get_unique_identifiers_flag(name)
         pages = self.get_total_pages(name, 250)
+        id_fields = {"id", "identifier"}
 
         @sleep_and_retry
         @limits(calls=175, period=60)
@@ -401,7 +507,7 @@ class Client(SwaggerClient):
                     "project": name,
                     "page": page,
                     "per_page": 250,
-                    "_fields": "id,identifier,structures,tables",
+                    "_fields": ",".join(id_fields | COMPONENTS)
                 },
             )
             setattr(future, "track_id", page)
@@ -419,7 +525,7 @@ class Client(SwaggerClient):
                         ret["ids"].add(contrib["id"])
                         ret["identifiers"].add(contrib["identifier"])
 
-                        for component in ["structures", "tables"]:
+                        for component in COMPONENTS:
                             md5s = set(d["md5"] for d in contrib[component])
                             ret[component] |= md5s
 
@@ -431,6 +537,15 @@ class Client(SwaggerClient):
 
         return ret
 
+    def get_number_contributions(self, **query):
+        """Retrieve total number of contributions for query
+
+        See `client.contributions.get_entries()` for keyword arguments used in query.
+        """
+        return self.contributions.get_entries(
+            _fields=["id"], _limit=1, **query
+        ).result()["total_count"]
+
     def submit_contributions(
         self,
         contributions,
@@ -442,6 +557,21 @@ class Client(SwaggerClient):
     ):
         """Convenience function to submit a list of contributions
 
+        Example for a single contribution dictionary:
+
+        {
+            "project": "sandbox",
+            "identifier": "mp-4",
+            "data": {
+                "a": "3 eV",
+                "b": {"c": "hello", "d": 3},
+                "d.e.f": "nest via dot-notation"
+            },
+            "structures": [<pymatgen Structure>, ...],
+            "tables": [<pandas DataFrame>, ...],
+            "attachments": [<pathlib.Path>, ...]
+        }
+
         Args:
             contributions (list): list of contribution dicts to submit
             skip_dupe_check (bool): skip check for duplicates of identifiers and components
@@ -450,6 +580,10 @@ class Client(SwaggerClient):
             per_page (int): number of contributions to submit in each chunk/request
             max_workers (int): number of parallel requests to use to submit chunk
         """
+        if not contributions or not isinstance(contributions, list):
+            print("Please provide list of contributions to submit.")
+            return
+
         tic = time.perf_counter()
 
         if max_workers > MAX_WORKERS:
@@ -457,50 +591,64 @@ class Client(SwaggerClient):
             print(f"max_workers reset to max {MAX_WORKERS}")
 
         # get existing contributions
-        print("get existing contributions ...")
         existing = defaultdict(set)
         existing["unique_identifiers"] = True
-        project_name = contributions[0]["project"]
+        project_names = list(set(c["project"] for c in contributions))
+        initial_total = self.get_number_contributions(project__in=project_names)
 
         if not skip_dupe_check:
-            existing = self.get_contributions(project_name)
+            print("get existing contributions ...")
+            existing = {
+                project_name: self.get_contributions(project_name)
+                for project_name in project_names
+            }
 
         # prepare contributions
         print("prepare contributions ...")
-        contribs = []
-        digests = defaultdict(set)
-        components = {"structures", "tables"}
+        contribs = defaultdict(list)
+        digests = {project_name: defaultdict(set) for project_name in project_names}
         fields = [
             comp
             for comp in self.swagger_spec.definitions.get(
                 "ContributionsSchema"
             )._properties.keys()
-            if comp not in components
+            if comp not in COMPONENTS
         ]
 
-        # TODO parallelize?
         for contrib in tqdm(contributions, leave=False):
+            project_name = contrib["project"]
             if (
-                existing["unique_identifiers"]
-                and contrib["identifier"] in existing["identifiers"]
+                existing[project_name]["unique_identifiers"]
+                and contrib["identifier"] in existing[project_name]["identifiers"]
             ):
                 continue
 
-            contribs.append({k: deepcopy(contrib[k]) for k in fields if k in contrib})
+            contribs[project_name].append({
+                k: deepcopy(
+                    unflatten(contrib[k], splitter="dot")
+                    if k == "data" else contrib[k]
+                )
+                for k in fields if k in contrib
+            })
 
-            for component in components:
-                contribs[-1][component] = []
+            for component in COMPONENTS:
+                contribs[project_name][-1][component] = []
                 elements = contrib.get(component, [])
                 nelems = len(elements)
 
+                if nelems > MAX_ELEMS:
+                    raise ValueError(f"Too many {component} ({nelems} > {MAX_ELEMS})!")
+
                 for idx, element in enumerate(elements):
                     is_structure = isinstance(element, Structure)
+                    is_table = isinstance(element, pd.DataFrame)
+                    is_attachment = isinstance(element, Path)
                     if component == "structures" and not is_structure:
-                        raise ValueError("Only accepting pymatgen Structure!")
-                    elif component == "tables" and not isinstance(
-                        element, pd.DataFrame
-                    ):
-                        raise ValueError("Only accepting pandas DataFrame!")
+                        raise ValueError(f"Only accepting pymatgen Structure for {component}!")
+                    elif component == "tables" and not is_table:
+                        raise ValueError(f"Only accepting pandas DataFrame for {component}!")
+                    elif component == "attachments" and not is_attachment:
+                        raise ValueError(f"Only accepting pathlib.Path for {component}!")
 
                     if is_structure:
                         dct = element.as_dict()
@@ -509,11 +657,29 @@ class Client(SwaggerClient):
 
                         if not dct.get("charge"):
                             del dct["charge"]
-                    else:
+                    elif is_table:
                         element.index = element.index.astype(str)
                         for col in element.columns:
                             element[col] = element[col].astype(str)
                         dct = element.to_dict(orient="split")
+                    elif is_attachment:
+                        kind = guess(str(element))
+
+                        if not isinstance(kind, SUPPORTED_FILETYPES):
+                            raise ValueError(
+                                f"{element.name} not supported. Use one of {SUPPORTED_MIMES}!"
+                            )
+
+                        content = element.read_bytes()
+                        size = len(content)
+
+                        if size > MAX_BYTES:
+                            raise ValueError(f"{element.name} too large ({size} > {MAX_BYTES})!")
+
+                        dct = {
+                            "mime": kind.mime,
+                            "content": b64encode(content).decode("utf-8")
+                        }
 
                     digest = get_md5(dct)
 
@@ -523,7 +689,7 @@ class Client(SwaggerClient):
                             c = element.composition
                             comp = c.get_integer_formula_and_factor()
                             dct["name"] = f"{comp[0]}-{idx}" if nelems > 1 else comp[0]
-                    else:
+                    elif is_table:
                         name = f"table-{idx}" if nelems > 1 else "table"
                         dct["name"] = element.attrs.get("name", name)
                         title = element.attrs.get("title", dct["name"])
@@ -537,23 +703,28 @@ class Client(SwaggerClient):
                             labels["variable"] = variable
 
                         dct["attrs"] = {"title": title, "labels": labels}
+                    elif is_attachment:
+                        dct["name"] = element.name
 
-                    dupe = digest in digests[component] or digest in existing[component]
+                    dupe = bool(
+                        digest in digests[project_name][component] or
+                        digest in existing[project_name][component]
+                    )
 
                     if not ignore_dupes and dupe:
-                        msg = f"Duplicate: {dct['name']}!"
+                        msg = f"Duplicate in {project_name}: {contrib['identifier']} {dct['name']}"
                         raise ValueError(msg)
 
                     if not dupe:
-                        digests[component].add(digest)
-                        contribs[-1][component].append(dct)
+                        digests[project_name][component].add(digest)
+                        contribs[project_name][-1][component].append(dct)
 
         # submit contributions
         if contribs:
             print("submit contributions ...")
             with FuturesSession(max_workers=max_workers) as session:
                 # bravado future doesn't work with concurrent.futures
-                ncontribs = len(contribs)
+                total = 0
                 headers = {"Content-Type": "application/json"}
                 headers.update(self.headers)
 
@@ -566,29 +737,40 @@ class Client(SwaggerClient):
                         data=json.dumps(chunk).encode("utf-8"),
                     )
 
-                while contribs:
-                    futures = [
-                        post_future(chunk) for chunk in chunks(contribs, n=per_page)
-                    ]
+                for project_name in project_names:
+                    ncontribs = len(contribs[project_name])
+                    total += ncontribs
 
-                    self._run_futures(futures, total=len(contribs))
-
-                    if existing["unique_identifiers"] and retry:
-                        existing = self.get_contributions(project_name)
-                        contribs = [
-                            c
-                            for c in contribs
-                            if c["identifier"] not in existing["identifiers"]
+                    while contribs[project_name]:
+                        futures = [
+                            post_future(chunk) for chunk in chunks(
+                                contribs[project_name], n=per_page
+                            )
                         ]
-                    else:
-                        contribs = []  # abort retrying
-                        if not existing["unique_identifiers"] and retry:
-                            print("Please resubmit failed contributions manually.")
+
+                        self._run_futures(futures, total=ncontribs)
+
+                        if existing[project_name]["unique_identifiers"] and retry:
+                            existing[project_name] = self.get_contributions(project_name)
+                            contribs[project_name] = [
+                                c for c in contribs[project_name]
+                                if c["identifier"] not in existing[project_name]["identifiers"]
+                            ]
+                        else:
+                            contribs[project_name] = []  # abort retrying
+                            if not existing[project_name]["unique_identifiers"] and retry:
+                                print("Please resubmit failed contributions manually.")
 
                 self.load()
+                final_total = self.get_number_contributions(project__in=project_names)
+                submitted_total = final_total - initial_total
                 toc = time.perf_counter()
                 dt = (toc - tic) / 60
-                print(f"It took {dt:.1f}min to submit {ncontribs} contributions.")
+                print(f"It took {dt:.1f}min to submit {submitted_total} contributions.")
+
+                if submitted_total < total:
+                    failed_total = total - submitted_total
+                    print(f"{failed_total} contributions failed to submit. Check errors/warnings.")
         else:
             print("Nothing to submit.")
 
@@ -614,8 +796,8 @@ class Client(SwaggerClient):
                         if "warning" in resp:
                             print(resp["warning"])
                     elif status != 502:
-                        warnings.warn(resp["error"])
+                        print(resp["error"][:10000] + "...")
                 elif status not in {503, 504}:
-                    warnings.warn(response.content.decode("utf-8"))
+                    print("ERROR", response.content.decode("utf-8"))
 
         return responses
